"""Utilities for serializing content."""

import datetime

from django.core.serializers.json import DjangoJSONEncoder
from django.utils.encoding import force_text
from django.utils.functional import Promise


class DjbletsJSONEncoder(DjangoJSONEncoder):
    """A JSON encoder that supports lazy strings, datetimes, and other objects.

    This is a specialization of
    :py:class:`~django.core.serializers.json.DjangoJSONEncoder` the does the
    following:

    * Evaluates strings translated with
      :py:func:`~django.utils.translation.ugettext_lazy` or
      :py:func:`~django.utils.translation.gettext_lazy` to real strings.

    * Removes the milliseconds and microseconds
      from :py:class:`datetimes <datetime.datetime>` (unless setting
      ``strip_datetime_ms=False`` when constructing the encoder). This is
      to help keep timestamps from appearing too new when compared against
      data coming from a MySQL database (which historically, and by default,
      chops off milliseconds).

    * Serializes Django :py:class:`models <django.db.models.base.Model>` with
      a ``to_json`` method via that method.
    """

    def __init__(self, strip_datetime_ms=True, *args, **kwargs):
        """Initialize the encoder.

        Args:
            strip_datetime_ms (bool, optional):
                Determines whether milliseconds should be stripped from a
                :py:class:`~datetime.datetime`. This is ``True`` by default,
                to preserve the old behavior of the encoder.
        """
        super(DjbletsJSONEncoder, self).__init__(*args, **kwargs)

        self.strip_datetime_ms = strip_datetime_ms

    def default(self, obj):
        """Encode the object into a JSON-compatible structure.

        Args:
            obj (object):
                The object to encode.

        Returns:
            object:
            A JSON-compatible structure (e.g., a :py:class:`dict`,
            :py:class:`list`, py:class:`unicode`, or :py:class:`bytes` object).
        """
        if isinstance(obj, Promise):
            # Handles initializing lazily created ugettext messages.
            return force_text(obj)
        elif isinstance(obj, datetime.datetime) and self.strip_datetime_ms:
            # This is like DjangoJSONEncoder's datetime encoding
            # implementation, except that it filters out the milliseconds
            # in addition to microseconds. This ensures consistency between
            # database-stored timestamps and serialized objects.
            r = obj.isoformat()

            if obj.microsecond:
                r = r[:19] + r[26:]

            if r.endswith('+00:00'):
                r = r[:-6] + 'Z'

            return r
        elif hasattr(obj, 'to_json') and callable(obj.to_json):
            return obj.to_json()

        return super(DjbletsJSONEncoder, self).default(obj)



"""A registry for managing avatar services."""

from __future__ import unicode_literals

import logging
import warnings

from django.utils.translation import ugettext_lazy as _

from djblets.avatars.errors import (AvatarServiceNotFoundError,
                                    DisabledServiceError)
from djblets.avatars.services.gravatar import GravatarService
from djblets.avatars.services.url import URLAvatarService
from djblets.avatars.settings import AvatarSettingsManager
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         ATTRIBUTE_REGISTERED, DEFAULT_ERRORS,
                                         NOT_REGISTERED, Registry, UNREGISTER)
from djblets.siteconfig.models import SiteConfiguration


DISABLED_SERVICE = 'disabled_service'
DISABLED_SERVICE_DEFAULT = 'disabled_service_default'
UNKNOWN_SERVICE_DEFAULT = 'unknown_service_default'
UNKNOWN_SERVICE_DISABLED = 'unknown_service_disabled'
UNKNOWN_SERVICE_ENABLED = 'unknown_service_enabled'


AVATAR_SERVICE_DEFAULT_ERRORS = DEFAULT_ERRORS.copy()
AVATAR_SERVICE_DEFAULT_ERRORS.update({
    ALREADY_REGISTERED: _(
        'Could not register avatar service %(item)s: This service is already '
        'registered.'
    ),
    ATTRIBUTE_REGISTERED: _(
        'Could not register avatar service %(attr_value)s: This service is '
        'already registered.'
    ),
    DISABLED_SERVICE: _(
        'Could not fetch instance of %(service_id)s service: This service is '
        'disabled.'
    ),
    DISABLED_SERVICE_DEFAULT: _(
        'Could not set the default service to %(service_id)s: This service is '
        'disabled.'
    ),
    NOT_REGISTERED: _(
        'Unknown avatar service %(attr_value)s: This service is not '
        'registered.'
    ),
    UNKNOWN_SERVICE_DEFAULT: _(
        'Could not set the default avatar service to %(service_id)s: This '
        'service is not registered.'
    ),
    UNKNOWN_SERVICE_DISABLED: _(
        'Could not disable unknown avatar service %(service_id)s: This '
        'service is not registered.'
    ),
    UNKNOWN_SERVICE_ENABLED: _(
        'Could not enable unknown avatar service %(service_id)s: This service '
        'is not registered.'
    ),
    UNREGISTER: _(
        'Could not unregister unknown avatar service %(item)s: This service '
        'is not registered.'
    ),
})


logger = logging.getLogger(__name__)


class AvatarServiceRegistry(Registry):
    """A registry for avatar services.

    This registry manages a set of avatar services (see
    :py:mod:`djblets.avatars.services.gravatar` for an example). The registries
    are saved to the database and require the use of the
    :py:mod:`djblets.siteconfig` app.

    .. versionchanged:: 1.0.3

       The avatar configuration is now retrieved from and immediately written
       to the current :py:class:`~djblets.siteconfig.models.SiteConfiguration`,
       in order to ensure that the list of enabled avatar services and the
       default service are never stale. This differs from 1.0.0 through 1.0.2,
       where the callers could make change to the local state without ever
       risking it being written to the database (which is generally not the
       desired behavior anyway).
    """

    #: The key name for the list of enabled services.
    ENABLED_SERVICES_KEY = 'avatars_enabled_services'

    #: The key name for the default service.
    DEFAULT_SERVICE_KEY = 'avatars_default_service'

    lookup_attrs = ('avatar_service_id',)

    default_errors = AVATAR_SERVICE_DEFAULT_ERRORS

    lookup_error_class = AvatarServiceNotFoundError

    #: The default avatar service classes.
    default_avatar_service_classes = [
        GravatarService,
        URLAvatarService,
    ]

    #: The settings manager for avatar services.
    #:
    #: This should be changed in a subclass as the default cannot get or set
    #: avatar service configurations.
    settings_manager_class = AvatarSettingsManager

    def __init__(self):
        """Initialize the avatar service registry."""
        super(AvatarServiceRegistry, self).__init__()

        self._instance_cache = {}

    def get_avatar_service(self, avatar_service_id):
        """Return an instance of the requested avatar service.

        The instance will be instantiated with the
        :py:attr:`settings_manager_class`.

        Args:
            avatar_service_id (unicode):
                The unique identifier for the avatar service.

        Returns:
            djblets.avatars.services.base.AvatarService:
            The requested avatar service, or ``None`` if not found.

        Raises:
            djblets.avatars.errors.DisabledServiceError:
                The requested service is disabled.
        """
        service = self._instance_cache.get(avatar_service_id)

        if service is None:
            try:
                service_cls = self.get('avatar_service_id', avatar_service_id)
            except self.lookup_error_class:
                service_cls = None

            if not service_cls:
                return None
        else:
            service_cls = type(service)

        # It's important to note that if the service instance was in the
        # cache before, that doesn't mean it should be considered enabled.
        # Another SiteConfiguration instance on another process/server
        # may have disabled it, and our SiteConfiguration may have picked
        # that up since this was added to the cache.
        if not self.is_enabled(service_cls):
            self._instance_cache.pop(avatar_service_id, None)

            raise DisabledServiceError(self.format_error(
                DISABLED_SERVICE,
                service_id=avatar_service_id))

        if service is None:
            service = self._instance_cache[avatar_service_id] = \
                service_cls(self.settings_manager_class)

        return service

    @property
    def configurable_services(self):
        """Yield the enabled service instances that have configuration forms.

        Yields:
            tuple:
            The enabled service instances that have configuration forms.
        """
        return (
            self.get_avatar_service(service.avatar_service_id)
            for service in self.enabled_services
            if service.is_configurable() and not service.hidden
        )

    @property
    def enabled_services(self):
        """Return the enabled services.

        Returns:
            set:
            The set of enabled avatar services, as
            :py:class:`djblets.avatars.service.AvatarService` instances.
        """
        siteconfig = SiteConfiguration.objects.get_current()
        services = set()

        for service_id in siteconfig.get(self.ENABLED_SERVICES_KEY):
            try:
                service = self.get('avatar_service_id', service_id)
            except self.lookup_error_class:
                service = None

            if service is not None:
                services.add(service)

        return services

    @enabled_services.setter
    def enabled_services(self, services):
        """Set the enabled services.

        If the default service would be disabled by setting the set of enabled
        services, the default service will be set to ``None``.

        This is equivalent to setting :py:attr:`enabled_services` with
        ``save=True``. It's considered deprecated in favor of that method.

        Args:
            services (set):
                The set of services to set as enabled. Each element must be a
                subclass of :py:class:`~djblets.avatars.service.AvatarService`.

        Raises:
            djblets.avatars.errors.AvatarServiceNotFoundError:
                This exception is raised when an unknown avatar service is
                enabled.
        """
        warnings.warn('Setting AvatarServiceRegistry.enabled_services is '
                      'deprecated in favor of calling '
                      'AvatarServiceRegistry.set_enabled_services.',
                      DeprecationWarning)

        self.set_enabled_services(services, save=True)

    def set_enabled_services(self, services, save=True):
        """Set the enabled services.

        If the default service would be disabled by setting the set of enabled
        services, the default service will be set to ``None``.

        Args:
            services (set):
                The set of services to set as enabled. Each element must be a
                subclass of :py:class:`~djblets.avatars.service.AvatarService`.

            save (bool, optional):
                Whether to save the settings after setting the list of
                enabled services. The default is to immediately save.

        Raises:
            djblets.avatars.errors.AvatarServiceNotFoundError:
                This exception is raised when an unknown avatar service is
                enabled.
        """
        new_service_ids = set()

        for service in services:
            if service in self:
                new_service_ids.add(service.avatar_service_id)
            else:
                raise self.lookup_error_class(self.format_error(
                    UNKNOWN_SERVICE_ENABLED,
                    service_id=service.avatar_service_id))

        siteconfig = SiteConfiguration.objects.get_current()
        cur_service_ids = \
            set(siteconfig.get(self.ENABLED_SERVICES_KEY))

        if new_service_ids == cur_service_ids:
            return

        to_enable = new_service_ids - cur_service_ids
        to_disable = cur_service_ids - new_service_ids

        for service_id in to_disable:
            self.disable_service_by_id(service_id, save=False)

        for service_id in to_enable:
            self.enable_service_by_id(service_id, save=False)

        default_service = self.default_service

        if (default_service is not None and
            not self.is_enabled(type(default_service))):
            self.set_default_service(None, save=False)

        if save:
            self.save()

    @property
    def default_service(self):
        """The default avatar service.

        Returns:
            djblets.avatars.services.AvatarService:
            The default avatar service, or ``None`` if there isn't one.
        """
        siteconfig = SiteConfiguration.objects.get_current()
        default_service_id = siteconfig.get(self.DEFAULT_SERVICE_KEY)

        if default_service_id is None:
            return None

        enabled_service_ids = siteconfig.get(self.ENABLED_SERVICES_KEY)

        if default_service_id not in enabled_service_ids:
            # The settings listed a default service that was no longer in
            # the list of enabled services. Warn about this, fix it, and
            # save the value in settings.
            logger.error(self.format_error(DISABLED_SERVICE_DEFAULT,
                                           service_id=default_service_id))

            # Set this manually in the SiteConfiguration instead of using
            # disable_service_by_id, to avoid infinitely recursing into this
            # property and that method, and also because it's just not
            # necessary to go through everything that function does.
            siteconfig.set(self.DEFAULT_SERVICE_KEY, None)

            return None

        return self.get_avatar_service(default_service_id)

    def set_default_service(self, service, save=True):
        """Set the default avatar service.

        Args:
            service (type):
                The avatar service to set as default.

            save (bool):
                Whether or not the registry should be saved to the database
                afterwards.

        Raises:
            djblets.avatars.errors.AvatarServiceNotFoundError:
                Raised if the service cannot be found.

            djblets.avatars.errors.DisabledServiceError:
                Raised if the service is not enabled.
        """
        if service is None:
            default_service_id = None
        elif service not in self:
            raise self.lookup_error_class(self.format_error(
                UNKNOWN_SERVICE_DEFAULT, service_id=service.avatar_service_id))
        elif not self.is_enabled(service):
            raise DisabledServiceError(self.format_error(
                DISABLED_SERVICE_DEFAULT,
                service_id=service.avatar_service_id))
        else:
            default_service_id = service.avatar_service_id

        siteconfig = SiteConfiguration.objects.get_current()

        if default_service_id != siteconfig.get(self.DEFAULT_SERVICE_KEY):
            siteconfig.set(self.DEFAULT_SERVICE_KEY, default_service_id)

            if save:
                self.save()

    def has_service(self, service_id):
        """Return whether or not the avatar service ID is registered.

        Args:
            service_id (unicode):
                The service's unique identifier.

        Returns:
            bool: Whether or not the service ID is registered.
        """
        try:
            return self.get('avatar_service_id', service_id) in self
        except self.lookup_error_class:
            return False

    def disable_service(self, service, save=True):
        """Disable an avatar service.

        This has no effect if the service is already disabled. If the default
        service becomes be disabled, it becomes ``None``.

        Args:
            service (type):
                The service to disable

            save (bool, optional):
                Whether or not the avatar service registry will be saved to
                the database after disabling the service. This defaults to
                ``True``.

        Raises:
            djblets.avatars.errors.AvatarServiceNotFoundError:
                This is raised if the service is not registered.
        """
        if service not in self:
            raise self.lookup_error_class(self.format_error(
                UNKNOWN_SERVICE_DISABLED,
                service_id=service.avatar_service_id))

        self.disable_service_by_id(service.avatar_service_id, save=save)

    def disable_service_by_id(self, service_id, save=True):
        """Disable an avatar service based on its ID.

        This allows for disabling services that may or may not be registered,
        for instance those that were previously added by an extension that is
        no longer available.

        This has no effect if the service is already disabled. If the default
        service becomes be disabled, it becomes ``None``.

        Args:
            service_id (unicode):
                The ID of the service to disable.

            save (bool, optional):
                Whether or not the avatar service registry will be saved to
                the database after disabling the service. This defaults to
                ``True``.

        Raises:
            djblets.avatars.errors.AvatarServiceNotFoundError:
                This is raised if the service is not registered.
        """
        if (self.default_service is not None and
            self.default_service.avatar_service_id == service_id):
            self.set_default_service(None, save=save)

        self._instance_cache.pop(service_id, None)

        siteconfig = SiteConfiguration.objects.get_current()
        enabled_service_ids = list(siteconfig.get(self.ENABLED_SERVICES_KEY))

        try:
            enabled_service_ids.remove(service_id)
        except ValueError:
            # This wasn't enabled in the stored site configuration, so don't
            # try saving now. We're done.
            pass
        finally:
            siteconfig.set(self.ENABLED_SERVICES_KEY, enabled_service_ids)

            if save:
                self.save()

    def enable_service(self, service, save=True):
        """Enable an avatar service.

        Args:
            service (type):
                The service to enable. This must be a subclass of
                :py:class:`~djblets.avatars.service.AvatarService`.

            save (bool, optional):
                Whether or not the avatar service registry will be saved to the
                database after enabling the service. This defaults to ``True``.

        Raises:
            djblets.avatars.errors.AvatarServiceNotFoundError:
                This is raised if the service is not registered.
        """
        if service not in self:
            raise self.lookup_error_class(self.format_error(
                UNKNOWN_SERVICE_ENABLED, service_id=service.avatar_service_id))

        self.enable_service_by_id(service.avatar_service_id, save=save)

    def enable_service_by_id(self, service_id, save=True):
        """Enable an avatar service.

        Args:
            service_id (unicode):
                The ID of the service to enable. Callers must take care to
                ensure this matches a service stored in the registry.

            save (bool, optional):
                Whether or not the avatar service registry will be saved to the
                database after enabling the service. This defaults to ``True``.

        Raises:
            djblets.avatars.errors.AvatarServiceNotFoundError:
                This is raised if the service is not registered.
        """
        siteconfig = SiteConfiguration.objects.get_current()
        enabled_service_ids = siteconfig.get(self.ENABLED_SERVICES_KEY)

        if service_id not in enabled_service_ids:
            siteconfig.set(self.ENABLED_SERVICES_KEY,
                           enabled_service_ids + [service_id])

            if save:
                self.save()

    def is_enabled(self, service):
        """Return whether or not the given avatar service is enabled.

        Args:
            service (type):
                The service to check. This must be a ubclass of
                :py:class:`~djblets.avatars.service.AvatarService`.

        Returns:
            bool:
            Whether or not the service ID is registered.
        """
        if service not in self:
            return

        siteconfig = SiteConfiguration.objects.get_current()
        enabled_service_ids = siteconfig.get(self.ENABLED_SERVICES_KEY)

        return service.avatar_service_id in enabled_service_ids

    def unregister(self, service):
        """Unregister an avatar service.

        Note that unregistering a service does not disable it. That must be
        done manually through :py:meth:`disable_service`. Disabling is a
        persistent operation that affects all server instances now and in the
        future, while unregistering may occur during the normal shutdown of a
        particular thread or process (especially if done through an extension).

        Args:
            service (type):
                The avatar service to unregister.

        Raises:
            djblets.avatars.errors.AvatarServiceNotFoundError:
                Raised if the specified service cannot be found.
        """
        self._instance_cache.pop(service.avatar_service_id, None)

        super(AvatarServiceRegistry, self).unregister(service)

    def get_defaults(self):
        """Yield the default avatar services.

        Subclasses should override the
        :py:attr:`default_avatar_service_classes` attribute instead of this in
        most cases.
        """
        for service_class in self.default_avatar_service_classes:
            yield service_class

    def save(self):
        """Save the avatar configuration to the database.

        As the avatar configuration is stored in the
        :py:class:`~djblets.siteconfig.models.SiteConfiguration`, this method
        will save any pending configuration, synchronizing it to all other
        processes/servers.

        If there are pending avatar configuration changes (due to passing
        ``save=False`` to some methods), and there's a separate call to
        :py:meth:`SiteConfiguration.save()
        <djblets.siteconfig.models.SiteConfiguration.save>` without calling
        this method, the new avatar configuration will still be saved.
        """
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.save()

    def for_user(self, user, service_id=None):
        """Return the requested avatar service for the given user.

        The following options will be tried:

        * The requested avatar service (if it is enabled)
        * The user's chosen avatar service (if it is enabled)
        * The default avatar service (which may be ``None``)

        Args:
            user (django.contrib.auth.models.User):
                The user to retrieve the avatar service for.

            service_id (unicode, optional):
                The unique identifier of the service that is to be retrieved.
                If this is ``None``, the default service will be used.

        Returns:
            djblets.avatars.services.base.AvatarService:
            An avatar service, or ``None`` if one could not be found.
        """
        settings_manager = self.settings_manager_class(user)
        user_service_id = settings_manager.avatar_service_id

        for sid in (service_id, user_service_id):
            if sid is None or not self.has_service(sid):
                continue

            service = self.get('avatar_service_id', sid)

            if self.is_enabled(service):
                return self.get_avatar_service(sid)

        return self.default_service


SiteConfiguration.add_global_defaults({
    AvatarServiceRegistry.ENABLED_SERVICES_KEY: [],
    AvatarServiceRegistry.DEFAULT_SERVICE_KEY: None,
})



"""Standard views for displaying and interacting with configuration forms."""

from __future__ import unicode_literals

import logging

from django.http import Http404, HttpResponseBadRequest, HttpResponseRedirect
from django.utils import six
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_protect
from django.views.generic.base import TemplateView


logger = logging.getLogger(__name__)


class ConfigPagesView(TemplateView):
    """Base view for a set of configuration pages.

    This will render the page for managing a set of configuration sub-pages
    (:py:class:`djblets.configforms.pages.ConfigPage` subclasses).

    Subclasses are expected to provide :py:attr:`title` and
    :py:attr:`page_classes`, at a minimum.

    To dynamically compute pages, implement a :py:attr:`page_classes` method
    and decorate it with ``@property``.
    """

    #: The page title.
    title = None

    #: The title to display over the navigation sidebar.
    nav_title = None

    #: The element ID for the page's container.
    pages_id = 'config_pages'

    #: The template used to render the view.
    template_name = 'configforms/config.html'

    #: The base template that the default template should inherit from.
    base_template_name = 'base.html'

    #: The list of page subclasses to display.
    page_classes = []

    #: The list of CSS bundle names to include on the page.
    css_bundle_names = []

    #: The list of JavaScript bundle names to include on the page.
    js_bundle_names = []

    #: The optional Backbone model used for the view state.
    js_model_class = None

    #: The Backbone view used to render the main view.
    js_view_class = 'Djblets.Config.PagesView'

    #: The HTTP methods that this view will respond to.
    http_method_names = ['get', 'post']

    @method_decorator(csrf_protect)
    def dispatch(self, request, *args, **kwargs):
        self.pages = [
            page_cls(self, request, request.user)
            for page_cls in self.page_classes
        ]

        self.forms = {}

        # Store a mapping of form IDs to form instances, and check for
        # duplicates.
        for page in self.pages:
            for form in page.forms:
                # This should already be handled during form registration.
                assert form.form_id not in self.forms, \
                    'Duplicate form ID %s (on page %s)' % (
                        form.form_id, page.page_id)

                self.forms[form.form_id] = form

        return super(ConfigPagesView, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """Handle HTTP post requests to the view.

        This will instantiate the form that the POST was targeting, validate
        the form, and then trigger its save if valid, reloading the page
        with the new state.

        Args:
            request (HttpRequest):
                The HTTP request from the client.

        Returns:
            HttpResponse:
                Either a redirect (if successfully posted), or a standard
                response with the page content.
        """
        form_id = request.POST.get('form_target')

        if form_id is None:
            logger.error('Submitted form content had no form_target.',
                         request=request)
            return HttpResponseBadRequest()

        if form_id not in self.forms:
            raise Http404

        # Replace the form in the list with a new instantiation containing
        # the form data. If we fail to save, this will ensure the error is
        # shown on the page.
        old_form = self.forms[form_id]
        page = old_form.page
        form_cls = old_form.__class__
        form = form_cls(page, request, request.user, request.POST)
        self.forms[form_id] = form

        # Replace the form in the page.
        for i, page_form in enumerate(page.forms):
            if page_form.form_id == form_id:
                page.forms[i] = form
                break

        if form.is_valid():
            form.save()

            return HttpResponseRedirect(request.path)

        # For our purposes, there's no difference between the handling of a
        # GET request and a POST request at this stage.
        return self.get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """Return extra context data to render in the template.

        Returns:
            dict: A dictionary of context data for the page.
        """
        return {
            'base_template_name': self.base_template_name,
            'page_title': self.title,
            'nav_title': self.nav_title or self.title,
            'pages_id': self.pages_id,
            'pages': self.pages,
            'css_bundle_names': self.css_bundle_names,
            'js_bundle_names': self.js_bundle_names,
            'js_model_class': self.js_model_class,
            'js_view_class': self.js_view_class,
            'js_model_data': self.get_js_model_data(),
            'js_view_data': self.get_js_view_data(),
            'forms': list(six.itervalues(self.forms)),
        }

    def get_js_view_data(self):
        """Return custom options to pass to the JavaScript view.

        By default, this will return an empty dictionary. Subclasses can
        override to provide custom data.

        Returns:
            dict:
                A dictionary of options to pass to the View instance. By
                default, this will be empty.
        """
        return {}

    def get_js_model_data(self):
        """Return custom attributes to pass to the JavaScript model.

        If :py:attr:`js_model_class` is provided, the data returned from this
        function will be provided to the model when constructued.

        By default, this will return an empty dictionary. Subclasses can
        override to provide custom data.

        Returns:
            dict:
                A dictionary of attributes to pass to the Model instance. By
                default, this will be empty.
        """
        return {}



#
# views.py -- Views for the authentication app
#
# Copyright (c) 2007-2009  Christian Hammond
# Copyright (c) 2007-2009  David Trowbridge
# Copyright (C) 2007 Micah Dowty
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
"""Authentication-related views.

The bundled views help with common authentication-related tasks not otherwise
provided by Django. At the moment, there is only support here for registration.
"""

from __future__ import unicode_literals

from django.contrib import auth
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.template.context import RequestContext
from django.views.decorators.csrf import csrf_protect

from djblets.auth.forms import RegistrationForm
from djblets.auth.signals import user_registered
from djblets.auth.util import validate_test_cookie


@csrf_protect
def register(request, next_page, form_class=RegistrationForm,
             extra_context=None, initial_values=None,
             form_kwargs=None, template_name="accounts/register.html"):
    """Handle registration of a new user.

    This works along with :py:class:`djblets.auth.forms.RegistrationForm`
    to register a new user. It will display a registration form, validate
    the user's new information, and then log them in.

    The registration form, next page, and context can all be customized by
    the caller.

    Args:
        request (HttpRequest):
            The HTTP request from the client.

        next_page (unicode):
            The URL to navigate to once registration is successful.

        form_class (Form subclass):
            The form that will handle registration, field validation, and
            creation of the user.

        extra_context (dict):
            Extra context variables to pass to the template when rendering.

        initial_values (dict):
            Initial values to set on the form when it is rendered.

        form_kwargs (dict):
            Additional keyword arguments to pass to the form class during
            instantiation.

        template_name (unicode):
            The name of the template containing the registration form.

    Returns:
        HttpResponse: The page's rendered response or redirect.
    """
    if initial_values is None:
        initial_values = {}

    if form_kwargs is None:
        form_kwargs = {}

    if request.method == 'POST':
        form = form_class(data=request.POST, request=request, **form_kwargs)
        form.full_clean()
        validate_test_cookie(form, request)

        if form.is_valid():
            user = form.save()
            if user:
                user = auth.authenticate(
                    username=form.cleaned_data['username'],
                    password=form.cleaned_data['password1'])
                assert user
                auth.login(request, user)
                try:
                    request.session.delete_test_cookie()
                except KeyError:
                    # Do nothing
                    pass

                # Other components can listen to this signal to
                # perform additional tasks when a new user registers
                user_registered.send(sender=None, user=request.user)

                return HttpResponseRedirect(next_page)
    else:
        form = form_class(initial=initial_values, request=request,
                          **form_kwargs)

    request.session.set_test_cookie()

    context = {
        'form': form,
    }

    if extra_context:
        context.update(extra_context)

    return render(request, template_name, context)



from __future__ import unicode_literals

from djblets.conditions.choices import (BaseConditionChoice,
                                        BaseConditionStringChoice,
                                        ConditionChoiceMatchListItemsMixin,
                                        ConditionChoices)
from djblets.conditions.errors import (ConditionChoiceNotFoundError,
                                       ConditionOperatorNotFoundError)
from djblets.conditions.operators import (BaseConditionOperator,
                                          ConditionOperators)
from djblets.testing.testcases import TestCase


class BaseConditionChoiceTests(TestCase):
    """Unit tests for djblets.conditions.choices.BaseConditionChoice."""

    def test_get_operator(self):
        """Testing BaseConditionChoice.get_operator"""
        class MyOperator(BaseConditionOperator):
            operator_id = 'my-op'

        class MyChoice(BaseConditionChoice):
            operators = ConditionOperators([
                MyOperator,
            ])

        choice = MyChoice()

        self.assertEqual(choice.get_operator('my-op').__class__, MyOperator)

    def test_get_operator_with_invalid_id(self):
        """Testing BaseConditionChoice.get_operator with invalid ID"""
        class MyChoice(BaseConditionChoice):
            operators = ConditionOperators()

        choice = MyChoice()

        with self.assertRaises(ConditionOperatorNotFoundError):
            choice.get_operator('invalid')

    def test_get_operators(self):
        """Testing BaseConditionChoice.get_operators"""
        class MyOperator1(BaseConditionOperator):
            operator_id = 'my-op-1'

        class MyOperator2(BaseConditionOperator):
            operator_id = 'my-op-2'

        class MyChoice(BaseConditionChoice):
            operators = ConditionOperators([
                MyOperator1,
                MyOperator2,
            ])

        choice = MyChoice()

        operators = list(choice.get_operators())
        self.assertEqual(len(operators), 2)
        self.assertEqual(operators[0].__class__, MyOperator1)
        self.assertEqual(operators[1].__class__, MyOperator2)


class ConditionChoiceMatchListItemsMixinTests(TestCase):
    """Unit tests for ConditionChoiceMatchListItemsMixin."""

    def test_matches_with_all_required_and_match(self):
        """Testing ConditionChoiceMatchListItemsMixin.matches with
        require_match_all_items=True and match
        """
        class MyChoice(ConditionChoiceMatchListItemsMixin,
                       BaseConditionStringChoice):
            require_match_all_items = True

        choice = MyChoice()

        self.assertTrue(choice.matches(
            operator=choice.get_operator('contains'),
            match_value=['foo1', 'foo2', 'foo3'],
            condition_value='foo',
            value_state_cache={}))

    def test_matches_with_all_required_and_no_match(self):
        """Testing ConditionChoiceMatchListItemsMixin.matches with
        require_match_all_items=True and no match
        """
        class MyChoice(ConditionChoiceMatchListItemsMixin,
                       BaseConditionStringChoice):
            require_match_all_items = True

        choice = MyChoice()

        self.assertFalse(choice.matches(
            operator=choice.get_operator('contains'),
            match_value=['foo1', 'foo2', 'bar1'],
            condition_value='foo',
            value_state_cache={}))

    def test_matches_with_any_required_and_match(self):
        """Testing ConditionChoiceMatchListItemsMixin.matches with
        require_match_all_items=False and match
        """
        class MyChoice(ConditionChoiceMatchListItemsMixin,
                       BaseConditionStringChoice):
            require_match_all_items = False

        choice = MyChoice()

        self.assertTrue(choice.matches(
            operator=choice.get_operator('contains'),
            match_value=['foo1', 'foo2', 'bar1', 'bar2'],
            condition_value='foo',
            value_state_cache={}))

    def test_matches_with_any_required_and_no_match(self):
        """Testing ConditionChoiceMatchListItemsMixin.matches with
        require_match_all_items=False and no match
        """
        class MyChoice(ConditionChoiceMatchListItemsMixin,
                       BaseConditionStringChoice):
            require_match_all_items = False

        choice = MyChoice()

        self.assertFalse(choice.matches(
            operator=choice.get_operator('contains'),
            match_value=['foo1', 'foo2'],
            condition_value='bar',
            value_state_cache={}))


class ConditionChoicesTests(TestCase):
    """Unit tests for djblets.conditions.choices.ConditionChoices."""

    def test_init_with_class_choices(self):
        """Testing ConditionChoices initialization with class-defined choices
        """
        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice-1'

        class MyChoices(ConditionChoices):
            choice_classes = [
                MyChoice1,
            ]

        choices = MyChoices()
        self.assertEqual(list(choices), [MyChoice1])

    def test_init_with_caller_choices(self):
        """Testing ConditionChoices initialization with caller-defined choices
        """
        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice-1'

        class MyChoice2(BaseConditionChoice):
            choice_id = 'my-choice-2'

        class MyChoices(ConditionChoices):
            choice_classes = [
                MyChoice1,
            ]

        choices = MyChoices([MyChoice2])
        self.assertEqual(list(choices), [MyChoice2])

    def test_get_choice(self):
        """Testing ConditionChoices.get_choice"""
        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice-1'

        choices = ConditionChoices([MyChoice1])
        self.assertEqual(choices.get_choice('my-choice-1').__class__,
                         MyChoice1)

    def test_get_choice_with_invalid_id(self):
        """Testing ConditionChoices.get_choice with invalid ID"""
        choices = ConditionChoices()

        with self.assertRaises(ConditionChoiceNotFoundError):
            choices.get_choice('invalid')

    def test_get_choice_with_kwargs(self):
        """Testing ConditionChoices.get_choice with kwargs for extra state"""
        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice-1'

        choices = ConditionChoices([MyChoice1])
        choice = choices.get_choice('my-choice-1', choice_kwargs={'abc': 123})
        self.assertEqual(choice.extra_state, {'abc': 123})

    def test_get_choices(self):
        """Testing ConditionChoices.get_choices"""
        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice-1'

        class MyChoice2(BaseConditionChoice):
            choice_id = 'my-choice-2'

        choices = ConditionChoices([MyChoice1, MyChoice2])

        choices = list(choices.get_choices())
        self.assertEqual(len(choices), 2)
        self.assertEqual(choices[0].__class__, MyChoice1)
        self.assertEqual(choices[1].__class__, MyChoice2)

    def test_get_choices_with_kwargs(self):
        """Testing ConditionChoices.get_choices with kwargs"""
        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice-1'

        class MyChoice2(BaseConditionChoice):
            choice_id = 'my-choice-2'

        choices = ConditionChoices([MyChoice1, MyChoice2])

        choices = list(choices.get_choices(choice_kwargs={'abc': 123}))
        self.assertEqual(len(choices), 2)
        self.assertEqual(choices[0].extra_state, {'abc': 123})
        self.assertEqual(choices[1].extra_state, {'abc': 123})



#
# gravatars.py -- Decorational template tags
#
# Copyright (c) 2008-2009  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from django import template
from django.utils.html import format_html

from djblets.gravatars import (get_gravatar_url,
                               get_gravatar_url_for_email)


register = template.Library()


@register.simple_tag
def gravatar(user, size=None):
    """Return HTML for displaying a user's Gravatar.

    This is also influenced by the following Django settings:

    ``GRAVATAR_SIZE``:
        The default size for Gravatars.

    ``GRAVATAR_RATING``:
        The maximum allowed rating (one of ``'g'``, ``'pg'``, ``'r'``, or
        ``'x'``).

    ``GRAVATAR_DEFAULT``:
        The default image to show if the user hasn't specified a Gravatar (one
        of ``identicon``, ``monsterid``, or ``wavatar``).

    See https://gravatar.com for more information.

    Args:
        user (django.contrib.auth.models.User):
            The user whose gravatar is to be displayed.

        size (int):
            An optional height and width for the image (in pixels). This will
            default to 80 if not specified.

    Returns:
        django.utils.safestring.SafeText:
        HTML for rendering the Gravatar.
    """
    url = get_gravatar_url(user=user, size=size)

    if url:
        return format_html(
            '<img src="{0}" width="{1}" height="{1}" alt="{2}" '
            'class="gravatar"/>',
            url, size, user.get_full_name() or user.username)
    else:
        return ''


@register.simple_tag
def gravatar_url(email, size=None):
    """Return a Gravatar URL for an e-mail address.

    This is also influenced by the following Django settings:

    ``GRAVATAR_SIZE``:
        The default size for Gravatars.

    ``GRAVATAR_RATING``:
        The maximum allowed rating (one of ``'g'``, ``'pg'``, ``'r'``, or
        ``'x'``).

    ``GRAVATAR_DEFAULT``:
        The default image to show if the user hasn't specified a Gravatar (one
        of ``identicon``, ``monsterid``, or ``wavatar``).

    See https://gravatar.com for more information.

    Args:
        email (unicode):
            The e-mail address.

        size (int):
            An optional height and width of the image (in pixels). This will
            default to 80 if not specified.

    Returns:
        django.utils.safestring.SafeText:
        HTML for rendering the Gravatar.
    """
    return get_gravatar_url_for_email(email=email, size=size)



"""Test case support for full API testing.

This provides a useful mixin, :py:class:`WebAPITestCaseMixin`, that can be
mixed into a base test case for an API testing suite. This mixin provides
a number of useful helpers for calling APIs and validating results, along
with generating consistent docstrings.
"""

from __future__ import print_function, unicode_literals

import json
import pprint

from django.test.client import BOUNDARY, MULTIPART_CONTENT, encode_multipart
from django.utils import six

from djblets.testing.testcases import TestCase


class WebAPITestCaseMixin(TestCase):
    """Mixin for test cases for thoroughly testing APIs.

    This helps with the testing of APIs by providing convenient methods for
    performing HTTP requests and checking their response payloads, redirects,
    mimetypes, and more.

    Any test methods in a subclass that have the ``is_test_template`` flag set
    to ``True`` can include ``<URL>`` in their docstring. When the test is
    run, the ``<URL>`` will be replaced by the contents of the subclass's
    :py:attr:`sample_api_url`.

    WebAPITestCaseMixin is meant to be mixed into a new base class that all
    API test suites would inherit from. Subclasses must provide values for
    :py:attr:`sample_api_url` and :py:attr:`error_mimetype`.
    """

    #: The path to the resource, which may be placed into the docstring.
    sample_api_url = None

    #: The mimetype representing an error for this API.
    error_mimetype = None

    #: A fake base URL for API operations.
    base_url = 'http://testserver'

    def shortDescription(self):
        """Return a description for a particular test.

        If the test has the ``is_test_template`` attribute set, and
        contains ``<URL>`` in the docstring, the URL will be replaced by
        the contents of :py:attr:`sample_api_url`.

        Returns:
            unicode: The description of the test.
        """
        desc = super(WebAPITestCaseMixin, self).shortDescription()

        if self.sample_api_url:
            test_method = getattr(self, self._testMethodName)

            if getattr(test_method, 'is_test_template', False):
                desc = desc.replace('<URL>', self.sample_api_url)

        return desc

    def api_get(self, path, query={}, follow_redirects=False,
                expected_status=200, expected_redirects=[],
                expected_headers={}, expected_mimetype=None,
                expected_json=True, **extra):
        """Perform and check a HTTP GET request, returning additional data.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            query (dict):
                The query string.

            expected_status (int):
                The expected HTTP status.

            follow_redirects (bool):
                Whether to expect and follow redirects to another URL.

            expected_redirects (list of unicode):
                The list of expected redirects performed by the resource(s),
                in order.

            expected_headers (dict):
                Expected HTTP headers and their values from the response.

            expected_mimetype (unicode):
                The expected mimetype for the response payload.

            expected_json (bool):
                Whether the response is expected to be in JSON format.

        Returns:
            dict or bytes:
            The parsed payload data as a dictionary, if ``expected_json`` is
            ``True`` and the response isn't a HTTP 302. Otherwise, the
            raw payload contents.
        """
        return self.api_get_with_response(path, query, follow_redirects,
                                          expected_status, expected_redirects,
                                          expected_headers, expected_mimetype,
                                          expected_json, **extra)[0]

    def api_get_with_response(self, path, query={}, follow_redirects=False,
                              expected_status=200, expected_redirects=[],
                              expected_headers={}, expected_mimetype=None,
                              expected_json=True, **extra):
        """Perform and check a HTTP GET request to an API resource.

        This will perform the request to the resource and validate
        that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            query (dict):
                The query string.

            expected_status (int):
                The expected HTTP status.

            follow_redirects (bool):
                Whether to expect and follow redirects to another URL.

            expected_redirects (list of unicode):
                The list of expected redirects performed by the
                resource(s),
                in order.

            expected_headers (dict):
                Expected HTTP headers and their values from the
                response.

            expected_mimetype (unicode):
                The expected mimetype for the response payload.

            expected_json (bool):
                Whether the response is expected to be in JSON format.

        Returns:
            tuple:
            A 2-tuple of:

            * the parsed payload (:py:class:`dict`, if ``expected_json`` is
              ``True`` or :py:class:`bytes`)
            * the raw response
              (:py:class:`djblets.webapi.responses.WebAPIResponse`).
        """
        path = self._normalize_path(path)

        print('GETing %s' % path)
        print("Query data: %s" % query)

        response = self._api_func_wrapper(
            self.client.get, path, query, expected_status, follow_redirects,
            expected_redirects, expected_mimetype,
            content_type='text/html; charset=utf-8',
            extra=extra)

        for header, value in six.iteritems(expected_headers):
            self.assertIn(header, response)
            self.assertEqual(response[header], value)

        if expected_status != 302 and expected_json:
            rsp = json.loads(response.content)
        else:
            rsp = response.content

        print('Response:')
        pprint.pprint(rsp)

        return rsp, response

    def api_post_with_response(self, path, query={}, expected_status=201,
                               expected_mimetype=None, **extra):
        """Perform an HTTP POST to an API resource, returning additional data.

        This works like :py:meth:`api_post`, but returns the resulting payload
        and the original :py:class:`HttpResponse`.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            query (dict):
                The form data to post.

            expected_status (int):
                The expected HTTP status.

            expected_mimetype (unicode):
                The expected mimetype for the response payload.

        Returns:
            tuple:
            A 2-item tuple containing the parsed response and the original
            :py:class:`HttpResponse`.
        """
        path = self._normalize_path(path)

        print('POSTing to %s' % path)
        print("Post data: %s" % query)
        response = self.client.post(path, query,
                                    HTTP_X_REQUESTED_WITH='XMLHttpRequest',
                                    **extra)
        print("Raw response: %s" % response.content)
        self.assertEqual(response.status_code, expected_status)

        if expected_status >= 400:
            self.assertEqual(expected_mimetype, None)

            if expected_status != 405:
                self.assertEqual(response['Content-Type'], self.error_mimetype)
        else:
            self.assertNotEqual(expected_mimetype, None)
            self.assertEqual(response['Content-Type'], expected_mimetype)

        return self._get_result(response, expected_status), response

    def api_post(self, *args, **kwargs):
        """Perform and check an HTTP POST to an API resource.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            query (dict):
                The form data to post.

            expected_status (int):
                The expected HTTP status.

            expected_mimetype (unicode):
                The expected mimetype for the response payload.

        Returns:
            The parsed payload data.
        """
        return self.api_post_with_response(*args, **kwargs)[0]

    def api_put(self, path, query={}, expected_status=200,
                follow_redirects=False, expected_redirects=[],
                expected_mimetype=None, **extra):
        """Perform and check an HTTP PUT to an API resource.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            query (dict):
                The form data to post.

            expected_status (int):
                The expected HTTP status.

            follow_redirects (bool):
                Whether to expect and follow redirects to another URL.

            expected_redirects (list of unicode):
                The list of expected redirects performed by the resource(s),
                in order.

            expected_mimetype (unicode):
                The expected mimetype for the response payload.

        Returns:
            The parsed payload data.
        """
        path = self._normalize_path(path)

        print('PUTing to %s' % path)
        print("Post data: %s" % query)
        data = encode_multipart(BOUNDARY, query)

        response = self._api_func_wrapper(
            self.client.put, path, data, expected_status, follow_redirects,
            expected_redirects, expected_mimetype,
            content_type=MULTIPART_CONTENT,
            extra=extra)

        return self._get_result(response, expected_status)

    def api_delete(self, path, expected_status=204, **extra):
        """Perform and check an HTTP DELETE to an API resource.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            expected_status (int):
                The expected HTTP status.

        Returns:
            HttpResponse:
            The HTTP respons of the payload. This won't have any content if
            the delete was successful.
        """
        path = self._normalize_path(path)

        print('DELETEing %s' % path)
        response = self.client.delete(path, **extra)
        print("Raw response: %s" % response.content)
        self.assertEqual(response.status_code, expected_status)

        return self._get_result(response, expected_status)

    def assertHttpOK(self, response, check_last_modified=False,
                     check_etag=False):
        """Assert that a response was OK and optionally has caching headers.

        Args:
            response (HttpResponse):
                The HTTP response from the API.

            check_last_modified (bool):
                If set, this will assert that the response has a
                ``Last-Modified`` header.

            check_etag (bool):
                If set, this will assert that the response has a
                ``ETag`` header.
        """
        self.assertEquals(response.status_code, 200)

        if check_last_modified:
            self.assertIn('Last-Modified', response)

        if check_etag:
            self.assertIn('ETag', response)

    def assertHttpNotModified(self, response):
        """Assert that the response was HTTP 304 Not Modified.

        Args:
            response (HttpResponse): The HTTP response from the API.
        """
        self.assertEquals(response.status_code, 304)
        self.assertEquals(response.content, '')

    def _api_func_wrapper(self, api_func, path, query, expected_status,
                          follow_redirects, expected_redirects,
                          expected_mimetype, content_type='', extra={}):
        """Wrapper around API functions.

        This will invoke an API function with all the necessary parameters,
        and check the results for the expected values.

        Args:
            api_func (callable):
                The API function to call.

            path (unicode):
                The path to the resource to request.

            query (dict):
                The GET query data or POST form data.

            expected_status (int):
                The expected HTTP status.

            follow_redirects (bool):
                Whether to expect and follow redirects to another URL.

            expected_redirects (list of unicode):
                The list of expected redirects performed by the resource(s),
                in order.

            expected_mimetype (unicode):
                The expected mimetype for the response payload.

            content_type (unicode):
                The content type of the request data.

            extra (dict):
                Extra data to pass to the function.

        Returns:
            HttpResponse: The HTTP response from the API resource.
        """
        response = api_func(path, query, follow=follow_redirects,
                            content_type=content_type,
                            HTTP_X_REQUESTED_WITH='XMLHttpRequest',
                            **extra)

        print("Raw response: %s" % response.content)

        self.assertEqual(response.status_code, expected_status)

        if expected_status >= 400:
            self.assertEqual(expected_mimetype, None)

            if expected_status != 405:
                self.assertEqual(response['Content-Type'], self.error_mimetype)
        elif expected_status != 302:
            self.assertNotEqual(expected_mimetype, None)
            self.assertEqual(response['Content-Type'], expected_mimetype)

        if expected_redirects:
            self.assertEqual(len(response.redirect_chain),
                             len(expected_redirects))

            for redirect in expected_redirects:
                self.assertEqual(response.redirect_chain[0][0],
                                 self.base_url + expected_redirects[0])

        return response

    def _normalize_path(self, path):
        """Normalize a path, removing the base URL if found.

        Args:
            path (unicode): The path or full URL.

        Returns:
            unicode: The path without the base server URL.
        """
        if path.startswith(self.base_url):
            return path[len(self.base_url):]
        else:
            return path

    def _get_result(self, response, expected_status):
        """Return the parsed result from a response.

        If the expected HTTP status is a 204 (No Content) or 405 (Method not
        allowed), the response will be asserted to be empty and the function
        will return ``None``. Otherwise, the parsed JSON response will be
        returned.

        Args:
            response (HttpResponse):
                The HTTP response from the API.

            expected_status (int):
                The expected HTTP status.

        Returns:
            The parsed JSON response from the payload, or ``None``.
        """
        if expected_status in (204, 405):
            self.assertEqual(response.content, '')
            rsp = None
        else:
            rsp = json.loads(response.content)
            print('Response:')
            pprint.pprint(rsp)

        return rsp



# -*- coding: utf-8 -*-
"""Tests for djblets.gravatars."""

from __future__ import unicode_literals

from hashlib import md5

from django.contrib.auth.models import User

from djblets.gravatars import get_gravatar_url, get_gravatar_url_for_email
from djblets.testing.testcases import TestCase


class GravatarTests(TestCase):
    """Tests for gravatar functions."""

    _URL_BASE = 'https://secure.gravatar.com/avatar/'

    def test_get_gravatar_url_for_email_uppercase(self):
        """Testing get_gravatar_url_for_email with uppercase characters"""
        self.assertEqual(get_gravatar_url_for_email(email='FOO@EXAMPLE.COM'),
                         get_gravatar_url_for_email(email='foo@example.com'))
        self.assertEqual(
            get_gravatar_url_for_email(email='foo@example.com'),
            '%s%s' % (self._URL_BASE, md5('foo@example.com').hexdigest()))

    def test_get_gravatar_url_for_email_whitespace(self):
        """Testing get_gravatar_url_for_email with whitespace characters"""
        self.assertEqual(get_gravatar_url_for_email(email=' foo@example.com '),
                         get_gravatar_url_for_email(email='foo@example.com'))

    def test_get_gravatar_url_for_email_unicode(self):
        """Testing get_gravatar_url_for_email with unicode characters"""
        raw_email = 'こんにちは@example.com'
        encoded_email = raw_email.encode('utf-8')

        self.assertEqual(get_gravatar_url_for_email(email=raw_email),
                         get_gravatar_url_for_email(email=encoded_email))
        self.assertEqual(
            get_gravatar_url_for_email(email=raw_email),
            '%s%s' % (self._URL_BASE, md5(encoded_email).hexdigest()))

    def test_get_gravatar_url_uppercase(self):
        """Testing get_gravatar_url with uppercase characters"""
        user = User.objects.create_user(username='user',
                                        email='FOO@EXAMPLE.COM')
        self.assertEqual(get_gravatar_url(user=user),
                         get_gravatar_url_for_email(email='foo@example.com'))

    def test_get_gravatar_url_whitespace(self):
        """Testing get_gravatar_url with whitespace characters"""
        user = User.objects.create_user(username='user',
                                        email=' foo@example.com ')
        self.assertEqual(get_gravatar_url(user=user),
                         get_gravatar_url_for_email(email='foo@example.com'))

    def test_get_gravatar_url_for_unicode(self):
        """Testing get_gravatar_url with unicode characters"""
        raw_email = 'こんにちは@example.com'
        encoded_email = raw_email.encode('utf-8')
        user = User.objects.create_user(username='user', email=raw_email)
        self.assertEqual(get_gravatar_url(user=user),
                         get_gravatar_url_for_email(email=encoded_email))
